﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using HeuristicLab.Algorithms.DataAnalysis.Experimental;
using HeuristicLab.Problems.DataAnalysis;

namespace Main {
  class Program {
    static void Main(string[] args) {
      var xs = HeuristicLab.Common.SequenceGenerator.GenerateSteps(-3.5, 3.5, 0.1, includeEnd: true).ToList();
      var ys = xs.Select(xi => 1.0 / Math.Sqrt(2 *Math.PI) * Math.Exp(-0.5 * xi*xi)).ToArray(); // 1.0 / (Math.Sqrt(2 * Math.PI) * Math.Exp(-0.5 * xi * xi))).ToArray();

      alglib.hqrndstate state;
      alglib.hqrndseed(1234, 5678, out state);
      var ys_noise = ys.Select(yi => yi + alglib.hqrndnormal(state) * 0.1).ToList();

      int n = xs.Count;
      double[] ys_smoothed = new double[n];
      double[] df = Enumerable.Repeat(1.0, n).ToArray();   // set each df if actual standard deviation of data points is not known
      double[,] c = new double[n - 1, 3];
      double var = -99.0;
      int ic = n - 1;
      int job = 1; // calc estimates of standard errors in points
      double[] se = new double[n]; // standard errors in points
      double[,] wk = new double[7, (n + 2)]; // work array;
      int ier = -99;
      if (Environment.Is64BitProcess) {
        // CubicSplineGCV.cubgcv_x64(xs.ToArray(), f, df, ref n, ys_noise.ToArray(), c, ref ic, ref var, ref job, se, wk);
        Console.WriteLine("x64 version not supported");
      } else {
        CubicSplineGCV.cubgcv_x86(xs.ToArray(), ys_noise.ToArray(), df, ref n, ys_smoothed,
          c, ref ic, ref var, ref job, se, wk, ref ier);


        //                           WK(1) = SMOOTHING PARAMETER (= RHO/(RHO + 1))
        //                           WK(2) = ESTIMATE OF THE NUMBER OF DEGREES OF
        //                                   FREEDOM OF THE RESIDUAL SUM OF SQUARES
        //                           WK(3) = GENERALIZED CROSS VALIDATION
        //                           WK(4) = MEAN SQUARE RESIDUAL
        //                           WK(5) = ESTIMATE OF THE TRUE MEAN SQUARE ERROR
        //                                   AT THE DATA POINTS
        //                           WK(6) = ESTIMATE OF THE ERROR VARIANCE
        //                           WK(7) = MEAN SQUARE VALUE OF THE DF(I)

        Console.WriteLine("Smoothing Parameter (= RHO/(RHO + 1) {0}", wk[0, 0]);
        Console.WriteLine("Estimated DOF of RSS {0}", wk[0, 1]);
        Console.WriteLine("GCV {0}", wk[0, 2]);
        Console.WriteLine("Mean squared residual {0}", wk[0, 3]);
        Console.WriteLine("Estimate of true MSE at data points {0}", wk[0, 4]);
        Console.WriteLine("Estimate of error variance {0}", wk[0, 5]);
        Console.WriteLine("Mean square value of DF(I) {0}", wk[0, 6]);

        OnlineCalculatorError error;
        var mse = OnlineMeanSquaredErrorCalculator.Calculate(ys, ys_smoothed, out error);
        Console.WriteLine("MSE(ys, ys_smooth) = {0}", mse);
        mse = OnlineMeanSquaredErrorCalculator.Calculate(ys, ys_noise, out error);
        Console.WriteLine("MSE(ys, ys_noise) = {0}", mse);

        Thread.CurrentThread.CurrentCulture = CultureInfo.InvariantCulture;

        for (int i=0;i<n;i++) {
          Console.WriteLine("{0}\t{1}\t{2}\t{3}\t{4}\t{5}", 
            xs[i], ys[i], ys_smoothed[i], ys_noise[i], ys_smoothed[i] + 1.96 * se[i], ys_smoothed[i] - 1.96 * se[i]);
        }
      }
    }
  }
}
